#ifndef RES_SHADER_H
#define RES_SHADER_H

#include <morfbase.h>
#include <glheaders.h>
#include "surfaceflags.h"
#include "shaderflags.h"
#include "font.h"
#include "texture.h"

#define MAX_LAYERS		8
#define MAX_DEFORMS		4
#define MAX_ANIMFRAMES	8
#define MAX_PARAMS		8

namespace FRONTEND
{

	class CResManagerGL;

	class CResShader : public CFEResource
	{
	public:
		typedef struct _funclist
		{
			int func;
			int parameters;
			float p[MAX_PARAMS];
			_funclist *next;
		} funclist;
		typedef struct
		{
			int			flags;					/**< layer flags */

			float		anim_speed;				/**< speed of the animation */
			int			num_maps;				/**< number of texures (for animated layers) */
			HResTexture	map[MAX_ANIMFRAMES];	/**< array of the maps */

			funclist	*tcMod;					/**< texture coordinates modifiers */
			funclist	*aGen;					/**< alpha color generation */

			TCGEN		tcGen;

			RGBGEN		rgbGen;
			float		rgbGenParams[5];

			ALPHAGEN	alphaGen;				/**< alpha channel generation */
			float		alphaGenParams[5];		/**< alpha channel parameter */

			int			alphafunc;				/**< alpha function */
			float		alphafuncref;

			int			blendsrc;				/**< blend function */
			int			blenddst;

			int			depthFunc;				/**< depth func */
			int			depthWrite;				/**< touch depth buffer */

			int frame;							/**< @todo  remove this */
		} Layer;
		typedef struct
		{
			Layer* layer;
			float params[5];
		} FogParms;
		typedef struct
		{
			int num_layers;				/**< 2 layers max */
			Layer* layer[2][6];			/**< layers used by farbox and nearbox */
			int cloudheight;			/**< apparent curvature of the cloud layers (default is 128) */
		} SkyParms;
		typedef struct
		{
			DEFORMVERTEXES		type;
			float				params[8];
		} DeformVertexes;

		typedef struct
		{
			HResTexture tex;
			GENERAL::TMorfPointf texCoords[4];

		} TElementInfo;

		typedef struct
		{
			long scale[2];
			HResTexture tex;
		} TTexInfo;

		enum EShaderStatus {
			SS_LOADED=0,
			SS_DEFAULT_LIGHTMAP,
			SS_DEFAULT,
			SS_DEFAULT_GUI,
			SS_NOTLOADED,};
	public:


		char	shadername[64];		/**< shader name - used for debug only */

		HResFont fnt;
		GENERAL::TColor4ub fntColor;

		int		surface_flags;		/**< surface_flags (described in surface_flags, come from the bsp, 
										and not from the shader definition) */
		int		content_flags;		/**< contents of the shader ? */
		int		cull;				/**< culling type */

		int		sortvalue;			/**< @todo Carmack changed int sortvalues to float, why? */


		std::map	<long, TTexInfo> textures;
		std::vector <TElementInfo>	elements;

		/**
		* Flags to determine misc properties of the shader, like polygonoffset,
		* multitexture, mipmap, nopicmip, haslightmap, portal... and to determine
		* which info of the datasource we need to render the surface
		*/
		int		flags;

		int		num_layers;			/**< number of layers */
		Layer *	layer[MAX_LAYERS];	/**< layers of stages (passes) table */

		int					num_deforms;					/**< number of deformations for the shader */
		DeformVertexes		deformVertexes[MAX_DEFORMS];	/**< vertex deformation */

		SkyParms			sky_params;						/**< sky parameters */
		FogParms			fog_params;						/**< fog parameters */

	protected:
		EShaderStatus shaderStatus;
	public:
		inline CResShader(CResourceManager *_resMan,char *_file, int _targetstatus) 
			: CFEResource(_resMan,_file,_targetstatus),	sortvalue(SORT_NORMAL) 
		{
			shadername[0]='\0';
			surface_flags=0;
			content_flags=0;
			cull=-1;
			sortvalue=-1;
			flags=0;
			num_layers=0;
			memset(layer,0,MAX_LAYERS*sizeof(Layer*));
			num_deforms=0;
			memset(deformVertexes,0,MAX_DEFORMS*sizeof(DeformVertexes));
			memset(&sky_params,0,sizeof(SkyParms));
			memset(&fog_params,0,sizeof(FogParms));
//SPTR			fnt = 0;
		};
		virtual ~CResShader();

		inline EShaderStatus GetShaderStatus() { return shaderStatus; };
		virtual bool LoadDefaultShader(char *_file) = 0;
		virtual bool LoadDefaultShaderLightMap(char *_file) = 0;
		virtual bool LoadDefaultShaderGUI(char *_file) = 0;


		Layer *CreateLayer();
	};

	/**
	* Parser class.
	* The parser class provides methods to parse files. This is specially used
	* for shaders parsing.
	*/
	#define MAXTOKEN			1024

	class CShaderFile
	{
	protected:
		char *fileName;
		char *fileContent;
		size_t fileSize;

	public:
		CShaderFile(char *_fileName);
		~CShaderFile();

		bool LoadShader(CResShader *shad,char *_shaderName);

	protected:
		char *GetShaderName();
		bool FindShader(char *_shaderName);
		bool SkipShader();

		void ParseLayer(CResShader *shad, CResShader::Layer *_layer);
		float ParseFunc(void);
	};

	class CShaderParser
	{
		public:
			static void StartParseBuffer(const char *buffer, const int size);
			static void StartParseString(const char *string);

			static void StopParseFile(void);

			static bool GetToken(bool crossline);
			static int GetOffset(void);
			static void GoTo(const int offset);
			static int IntForConst(const char *name);


			static char token[MAXTOKEN];				///< buffer for tokens 
			static int scriptline;

		protected:
			static CShaderFile *file;

			static const char *buffer;
			static const char *buf_start;
			static const char *buf_end;
	};

	class CResShaderGL : public CResShader
	{
	public:
		CResShaderGL(CResManagerGL *_resMan,char *_file, int _targetstatus);
		virtual ~CResShaderGL();

		virtual bool LoadDefaultShader(char *_file);
		virtual bool LoadDefaultShaderLightMap(char *_file);
		virtual bool LoadDefaultShaderGUI(char *_file);

		virtual bool LoadFromFile(char *_file, int _targetstatus);
		virtual bool LoadIntoGraphicAdapter(int _targetsatus);
		virtual bool UnloadFromGraphicAdapter();
	};

	typedef sptr<CResShader> HResShader;



} // namespace

#endif